import java.util.concurrent.TimeUnit;

public class OralTrainerImpl
{
  public static def Run(def api, def waitFxn)
  {
    m_Api = api;
    m_WaitFxn = waitFxn;

    LoadUserData();
    if (m_IsDebug)
    {
      boolean resume = DebugMode();
      if (!resume)
      {
        return "intro";
      }
    }

    if (IsFirstTime())
    {
      m_LastSession = m_Api.getTime();
      m_NextOrgasm = 100;
      IntroSequence();
    }
    else
    {
      ReturnSequence();
      m_LastSession = m_Api.getTime();
    }

    StoreUserData();

    GameSequence();

    OrgasmSequence();
    
    StoreUserData();
    
    return "intro";
  }

  private static void DommeSays(String message, int waitTime, boolean keepPicture, boolean allowClickThrough)
  {
    if (!keepPicture)
    {
      // TODO: Male avatar based on user preference
      int pictureNumber = m_Api.getRandom(5);
      while (pictureNumber == m_LastPicture)
      {
        pictureNumber = m_Api.getRandom(5);
      }

      String pictureUrl = String.format("OralTrainer/mistress/%02d.jpg", pictureNumber);
      m_Api.setImage(pictureUrl);
      m_LastPicture = pictureNumber;
    }

    m_Api.show(message);
    if (waitTime > 0)
    {
      if (allowClickThrough)
      {
         m_Api.showButton("(Skip)", waitTime);
      }
      else
      {
        Wait(waitTime, false);
      }
    }
  }

  private static void Wait(int waitTime, boolean showGauge)
  {
    if (showGauge)
    {
      m_Api.waitWithGauge(waitTime); // seconds
    }
    else
    {
      m_WaitFxn((float)waitTime);
      // TimeUnit.SECONDS.sleep(waitTime); // seconds
    }
  }

  private static void ShortWait(int waitTime)
  {
    TimeUnit.MILLISECONDS.sleep(waitTime); // milliseconds
  }

  private static void LoadUserData()
  {
    m_LastSession = 0;
    m_CurrentScore = 0;
    m_NextOrgasm = 0;
    m_EnduranceRating = 0;
    m_SpeedRating = 0;

    m_IsMale = !m_Api.loadBoolean("intro.female");
    m_Name = m_Api.loadString("intro.name");

    if (m_Api.loadString("OralTrainer.NextOrgasm"))
    {
      m_NextOrgasm = m_Api.loadInteger("OralTrainer.NextOrgasm");
    }

    if (m_Api.loadString("OralTrainer.CurrentScore"))
    {
      m_CurrentScore = m_Api.loadInteger("OralTrainer.CurrentScore");
    }

    if (m_Api.loadString("OralTrainer.EnduranceRating"))
    {
      m_EnduranceRating = m_Api.loadInteger("OralTrainer.EnduranceRating");
    }

    if (m_Api.loadString("OralTrainer.LickRating"))
    {
      m_EnduranceRating = m_Api.loadInteger("OralTrainer.LickRating");
    }

    if (m_Api.loadString("OralTrainer.SpeedRating"))
    {
      m_SpeedRating = m_Api.loadInteger("OralTrainer.SpeedRating");
    }

    if (m_Api.loadString("OralTrainer.LastSession"))
    {
      m_LastSession = m_Api.loadInteger("OralTrainer.LastSession");
    }

    if (m_Api.loadString("OralTrainer.IsDebug"))
    {
      m_IsDebug = m_Api.loadBoolean("OralTrainer.IsDebug");
    }
  }

  private static void ClearUserData()
  {
    m_LastSession = 0;
    m_CurrentScore = 0;
    m_NextOrgasm = 0;
    m_EnduranceRating = 0;
    m_SpeedRating = 0;
    m_LickRating = 0;

    StoreUserData();
  }

  private static void StoreUserData()
  {
    m_Api.save("OralTrainer.LastSession", m_LastSession);
    m_Api.save("OralTrainer.NextOrgasm", m_NextOrgasm);
    m_Api.save("OralTrainer.CurrentScore", m_CurrentScore);
    m_Api.save("OralTrainer.EnduranceRating", m_EnduranceRating);
    m_Api.save("OralTrainer.SpeedRating", m_SpeedRating);
    m_Api.save("OralTrainer.LickRating", m_LickRating);
  }

  private static boolean IsFirstTime()
  {
    return m_LastSession < 1;
  }

  private static void IntroSequence()
  {
    DommeSays("Hello, " + m_Name + ".", 3, false, true);

    DommeSays("You've been a bad " + GetPronoun() + " thinking you can please yourself as you desire.", 5, false, true);

    DommeSays("But that stops now.", 3, false, true);

    DommeSays("From now on you will earn every single orgasm.\n\n" +
      "With your mouth.", 6, false, true);

    WaitForToys();

    DisplayRules();
  }

  private static void ReturnSequence()
  {
    DommeSays("Welcome back, " + m_Name + ".", 3, false, true);

    def options = [
      "no",
      "yes"
    ];

    DommeSays("", 0, false, false);
    int selection = m_Api.getSelectedValue("Do you remember how this works?", options);
    if (selection == 0)
    {
        int delta = (int)(0.10 * m_NextOrgasm);
        m_NextOrgasm += delta;

        DommeSays("How disappointing. I'm moving your next orgasm back by " + delta + " points for wasting my time.", 5, false, true);
        
        WaitForToys();
        DisplayRules();
    }

    long time = m_Api.getTime();
    long delta = Math.ceil((time - m_LastSession) / 86400); // convert to days

    DommeSays("", 0, false, false);
    int orgasms = m_Api.getInteger("How many orgasms have you had since our last meeting?", 0);
    if (orgasms < 0)
    {
      orgasms = (delta + 1);
      DommeSays("Very funny. Now I'm going to be funny and assume " + orgasms + ".", 5, false, true);
    }
    else if (orgasms < Math.floor(delta / 60.0))
    {
      orgasms =  Math.ceil(delta / 60.0);
      DommeSays("I dont believe you. I'm going to assume " + orgasms + ".", 5, false, true);
    }

    if (orgasms > 0)
    {
      DommeSays("We'll just push your next orgasm back accordingly.", 5, false, true);

      for (int i=0; i<orgasms; ++i)
      {
        m_NextOrgasm += 100;
      }
    }   

    DommeSays("Your next orgasm is at " + m_NextOrgasm + " points. You have " + m_CurrentScore + " points carried over from previous sessions.", 5, false, true);
  }

  private static void GameSequence()
  {
    DommeSays("", 0, false, false);
    long sessionLength = m_Api.getInteger("How much time do you have to play? (minutes)", 10);

    if (sessionLength < 1)
    {
      DommeSays("You're funny. Come back when you're serious.", 5, false, true);
      return;
    }
    if (sessionLength < 5)
    {
      DommeSays("Just a little quickie then? I hope you weren't planning on an orgasm!", 6, false, true);
    }
    else if (sessionLength > 60)
    {
      sessionLength = 60;
      DommeSays("I appreciate your devotion, but I'm cutting you off at 60 minutes.", 5, false, true);
      DommeSays("We can't have you compensating for poor skills with long session times.", 6, false, true);
    }

    DommeSays("", 0, false, false);
    m_Api.getSelectedValue("Get ready (cuffs, chastity, and dildo), and prepare to worship cock.", ["Ready!"]);

    long endTime = m_Api.getTime() + sessionLength * 60;

    int lastGame = -1;
    while (m_Api.getTime() < endTime)
    {
      int remainder = Math.max(Math.ceil((endTime - m_Api.getTime()) / 60.0), 0);
      DommeSays("You have " +  m_CurrentScore + " of the " + m_NextOrgasm + " points you need for your next orgasm. You have " +
                remainder + " minutes remaining in this session.", 5, false, true);

      int game = m_Api.getRandom(4) % 4;
      while (lastGame == game)
      {
        game = m_Api.getRandom(4) % 4;
      }
      lastGame = game;

      if (game == 0)
      {
        EnduranceSession();
      }
      else if (game == 1)
      {
        SpeedSession();
      }
      else if (game == 2)
      {
        SkillSession();
      }
      else
      {
        LickSession();
      }
    }
  }

  private static void OrgasmSequence()
  {
    int punishment = -1;
    if (m_CurrentScore < m_NextOrgasm)
    {
      def options = [
        "Not really, no.", // 0
        "Please Mistress! I'll ruin my orgasm!", // 1
        "Please Mistress! I'll be less than a minute!", // 2
        "Please Mistress! I'll do it with one finger!", // 3
        "Please Mistress! I won't use my hands at all!", // 4
        "Please Mistress! I'll do it anally!", // 5
        "Please Mistress! I'll deepthroat the dildo until I cum!", // 6
        "Please Mistress! I'll fuck my throat until I cum!", // 7
      ];

      int adjOrgasm = m_NextOrgasm;
      DommeSays("", 0, false, false);
      punishment = m_Api.getSelectedValue("Uh-oh! You dont have enough points for an orgasm. Perhaps you'd like to beg a little?", options);
      switch (punishment)
      {
      case 0:
        m_NextOrgasm += 10;
        DommeSays("You must not be desperate enough then. Lets push your next orgasm back a bit.", 6, false, true);
        break;
      case 1:
        adjOrgasm *= 0.95;
        DommeSays("Hmmm... I do like the idea of taking your points away AND leaving you unfulfilled...", 6, false, true);
        break;
      case 2:
        adjOrgasm *= 0.9;
        if (m_IsMale)
        {
           DommeSays("Mmmm... We'll train you to cum prematurely in addition to being a cocksucker. How humiliating!", 7, false, true);
        }
        else
        {
           DommeSays("Hmmm... A wonton slut with epic blowjob skils. I like the possibilities...", 7, false, true);
        }
        break;
      case 3:
        adjOrgasm *= 0.85;
        DommeSays("Hmmm... just the pinky finger of your non-dominant hand...", 6, false, true);
        break;
      case 4:
        adjOrgasm *= 0.8;
        if (m_IsMale)
        {
           DommeSays("Cumming by rubbing your cock against that slick dildo. How humiliating!", 6, false, true);
        }
        else
        {
           DommeSays("Hmmm... no putting in your pussy either, of course...", 6, false, true);
        }
        break;
      case 5:
        adjOrgasm *= 0.7;
        DommeSays("Ohhh... punishing your ass AND no hands...", 6, false, true);
        break;
      case 6:
        adjOrgasm *= 0.6;
        DommeSays("Hmmm... maybe if you promise to stroke or rub only while deepthroating...", 6, false, true);
        break;
      case 7:
        adjOrgasm *= 0.5;
        DommeSays("Hmmm... perhaps if you stroke or rub in time with your throat fucking...", 6, false, true);
        break;
      default:
        DommeSays("What are you even talking about?", 3, false, true);
        break;
      }

      if (m_CurrentScore < adjOrgasm)
      {
        DommeSays("Not good enough " + m_Name + ". No orgasm today. Better try harder next time!", 6, false, true);
        return;
      }
    }

    if (punishment < 0)
    {
      DommeSays("Good " + GetPronoun() + ", " + m_Name + ". You have earned yourself an orgasm!", 3, false, true);
    }
    else
    {
      DommeSays("Good " + GetPronoun() + ", " + m_Name + ". You have earned yourself an orgasm! With caveats...", 3, false, true);
    }

    int maxIteration = 24 + m_Api.getRandom(36); // 2 - 5 minutes
    switch (punishment)
    {
    // case 0: No special action
    case 1:
      DommeSays("You may unlock yourself from chastity, kneel in front of me, and worship me. Stop rubbing/stroking as soon as you reach the point of no return.", 6, false, true);
      break;
    case 2:
      maxIteration = 8 + m_Api.getRandom(4); // 40 - 60 seconds
      DommeSays("You may unlock yourself from chastity, kneel in front of me, and worship me. You have " + (maxIteration * 5) + " seconds.", 6, false, true);
      break;
    case 3:
      DommeSays("You may unlock yourself from chastity, kneel in front of me, and worship me using just the pinky of your non-dominant hand.", 6, false, true);
      break;
    case 4:
      if (m_IsMale)
      {
        DommeSays("You may unlock yourself from chastity and position yourself so your cock touches the dildo.", 5, false, true);
        DommeSays("Rub yourself against the dildo until you cum, or until I tell you to stop.", 5, false, true);
        DommeSays("No hands. We can't have you cheating.", 3, false, true);
      }
      else
      {
        DommeSays("You may unlock yourself from chastity and position yourself so you clit presses agaisnt the dildo.", 5, false, true);
        DommeSays("Rub yourself against the dildo until you cum, or until I tell you to stop.", 5, false, true);
        DommeSays("No hands, and no taking the dildo in your pussy.", 3, false, true);
      }
      break;
    case 5:
      DommeSays("Lube the dildo, straddle it, and fuck your ass until you cum. Or until I get bored.", 5, false, true);
      DommeSays("No hands, and leave your chastity device locked in place. We can't have you cheating.", 5, false, true);
      break;
    case 6:
      DommeSays("You may unlock yourself from chastity, take the dildo all the way into your mouth, and start rubbing or stroking.", 6, false, true);
      DommeSays("You get one shot at this... as soon as you come up you're done.", 5, false, true);
      break;
    case 7:
      DommeSays("You may unlock yourself from chastity, and start thrusing the dildo in and out of your mouth.", 6, false, true);
      if (m_IsMale)
      {
        DommeSays("Pretend like you're sucking your own cock. Your hands follow your lips.", 5, false, true);
        DommeSays("That includes when your lips are off the dildo.", 5, false, true);
      }
      else
      {
        DommeSays("Rub your clit in time with your lips.", 5, false, true);
        DommeSays("When your lips are off the dildo, your hands leave your clit.", 5, false, true);
      }
      break;
    default:
      DommeSays("You may unlock yourself from chastity, kneel in front of me, and worship me until you cum. Or until I tell you to stop.", 6, false, true);
      break;
    }    

    DommeSays("Tell me when you're done.", 3, false, true);
    for (int i=0; i<maxIteration; ++i)
    {
      int pictureNumber = m_Api.getRandom(20);
      String pictureUrl = String.format("OralTrainer/mistress/orgasm%02d.jpg", pictureNumber);
      m_Api.setImage(pictureUrl);
        
      int result = 5;
      if (punishment == 6)
      {
        result = m_Api.showButton("*Gasp*", 5);
      }
      else
      {
        result = m_Api.showButton("Ahhhh", 5);
      }

      if (result < 5)
      {
        break;
      }

      if (i == maxIteration - 1)
      {
        DommeSays("Too slow! Hands off!", 3, false, true);
        DommeSays("BUt I'm still counting that as an attempted orgasm.", 3, false, true);
      }
    }

    DommeSays("Now, be a good " + GetPronoun() + " and lock yourself back in your chastity device.", 5, false, true);

    m_NextOrgasm += 100;
    m_CurrentScore = 0;
    DommeSays(String.format("Next time I wont go so easy on you. In fact, you'll %d points and you're starting at zero.", m_NextOrgasm), 3, false, true);
  }

  private static void WaitForToys()
  {
     DommeSays("Lets collect some toys for our session:", 3, false, true);

     // TODO: Image of dildo on chair
     m_Api.setImage("OralTrainer/toys/dildo.jpg");
     DommeSays("The only required toy is a dildo. Solidly affixed to a chair (via. suction, harness... whatever)", 4, true, true);

     DommeSays("Pick one that you could reasonably expect to deepthroat... 6-8 inches would be about right.", 4, true, true);

     m_Api.setImage("OralTrainer/toys/cuffs.jpg");
     DommeSays("Cuffs would also be a nice addition. I do so like the idea of you in bondage.", 4, true, true);

     DommeSays("Wrist, handcuffs, or rope... it doesn't matter as long as they keep your hands away from your mouth and naughty bits.", 4, true, true);

     DommeSays("Finally, I strongly recommend some way to keep you chaste between sessions.", 5, false, true);
     if (m_IsMale)
     {
        m_Api.setImage("OralTrainer/toys/mchas.jpg");

        DommeSays("I'm a fan of the 'Pet Trap' from Mature Metal. It's nearly indestructable.", 4, true, true);

        DommeSays("Though there are plenty of options for boys.", 4, true, true);
     }
     else
     {
       
        DommeSays("Sadly, boys have it easier here.", 3, false, true);

        m_Api.setImage("OralTrainer/toys/fchas.jpg");
        DommeSays("Though the 'Hybrid Chastity Belt' from Fancy Steel isn't a bad option.", 4, true, true);
     }

     DommeSays("If you can't find a 24/7 device, at least pick something you can wear overnight. Thats when you're most likely to cheat on me.", 6, false, true);

     DommeSays("Let me know when you're ready.", 0, false, true);

     m_Api.showButton("Ready Mistress!");
  }

  private static void DisplayRules()
  {
    DommeSays("Okay, " + m_Name + ". The rules of this game are simple:\n\n" + 
               "    1.) Go down when I tell you, as far as you can, and hold until I tell you to come up.\n\n" +
               "    2.) From now on, you only cum when and how I tell you to. No exceptions.", 10, false, true);

    DommeSays("Play proceeds in rounds from which you earn points towards your next orgasm. You select the length of the session. I select the games we play.\n\n"+
               "    - In endurance rounds your goal deepthroat the dildo until I tell you to come up.\n\n" +
               "    - In speed rounds you will fuck your throat, hard and fast, to my beat.\n\n" +
               "    - If I'm feeling generous, I'll let you merely lick the dildo in a pattern I describe.", 16, true, true);

    DommeSays("You will be asked to grade your performance after each game. Your feedback will adjust the difficulty of future sessions.", 6, false, true);

    DommeSays("To keep you from getting complacent, I increase the required number of points after every orgasm and set your score to 0.", 6, false, true);

    DommeSays("Its a bit of a losing game really.", 3, false, true);

    DommeSays("Unless you show vast improvement every session, you'll always have to increase the length of the session.", 6, false, true);

    DommeSays("And someday, even that won't be enough.", 0, false, false);

    m_Api.showButton("Got it!");
  }

  private static float EnduranceTest()
  {
    DommeSays("First, we need to get some baseline measurements. Don't worry- I'll remember you next time and we'll skip right to the juicy part.", 6, false, true);

    DommeSays("On my mark, take the dildo as deep into your throat as possible and hold it there. Tell me when you have to come up for air.", 6, false, true);

    DommeSays("Open Wide!", 0, false, false);
    m_Api.playSound("OralTrainer/open.mp3");

    Wait(2, false); 

    DommeSays("Down!", 0, false, false);
    m_Api.playSound("OralTrainer/down.mp3");
    int score = m_Api.showButton("*Gasp*");

    if (score > 900)
    {
      DommeSays("That is a really long time; I think you're lying to me.", 3, false, true);
      DommeSays("But the beauty of this game is that your lying result is its own punishment.", 3, false, true);
      DommeSays("Good luck earning enough points to cum.", 3, false, true);
    }
    else if (score < 15)
    {
      DommeSays("Pathetic. Either you're not trying, or you need some serious training.", 3, false, true);
      DommeSays("But the beauty of this game is that you'll have to improve a lot if you want to cum.", 3, false, true);
      DommeSays("Good luck earning enough points to cum.", 3, false, true);
    }
    else
    {
      DommeSays("Not bad. But I know that you can do even better.", 3, false, true);
      DommeSays("Don't worry... you'll be making pornstars jealous in no time.", 5, false, true);
    }

    m_EnduranceRating = score;
  }

  private static float SpeedTest()
  {
    DommeSays("First, We need to get some baseline measurements. Don't worry- I'll remember you next time and we'll skip right to the juicy part.", 6, false, true);

    DommeSays("On my mark, I want you to start thrusting the dildo in and out of your mouth. All the way in and all the way out. Repeat as many times as you can until I tell you to stop.", 8, false, true);

    DommeSays("Open Wide!", 0, false, false);
    m_Api.playSound("OralTrainer/open.mp3");

    Wait(2, false); 

    DommeSays("Go!", 0, false, false);
    m_Api.playSound("OralTrainer/down.mp3");

    m_Api.playBackgroundSound("OralTrainer/wait.mp3");
    Wait(60, true);

    DommeSays("Stop!", 0, false, false);
    m_Api.playSound("OralTrainer/up.mp3");
    
    DommeSays("", 0, false, false);
    int score = m_Api.getInteger("How many did you get?", 0);

    if (score > 120)
    {
      DommeSays("Thats over two thrusts per second! I think you're lying to me.", 3, false, true);
      DommeSays("But the beauty of this game is that your lying result is its own punishment.", 3, false, true);
      DommeSays("Good luck earning enough points to cum.", 3, false, true);
    }
    else if (score < 10)
    {
      DommeSays("Pathetic. Either you're not trying, or you need some serious training.", 3, false, true);
      DommeSays("But the beauty of this game is that you'll have to improve a lot if you want to cum.", 3, false, true);
      DommeSays("Good luck earning enough points to cum.", 3, false, true);
    }
    else
    {
      DommeSays("Not bad. But I know that you can do even better.", 3, false, true);
      DommeSays("Don't worry... we'll get that gag reflex licked in no time.", 5, false, true);
    }

     m_SpeedRating = score;
  }

  private static void EnduranceSession()
  {
    DommeSays("Lets work on your endurance.", 5, false, true);

    if (m_EnduranceRating < 1)
    {
      EnduranceTest();
    }

    DommeSays("Open Wide!", 0, false, false);
    m_Api.playSound("OralTrainer/open.mp3");

    DommeSays("Put it deep, deep, deep, down your throat.", 0, false, false);
    m_Api.playSound("OralTrainer/encourage00.mp3");

    long target = m_Api.getTime() + 3 * m_EnduranceRating * m_Warmup;
    if (m_Warmup < 0.99)
    {
      m_Warmup += 0.25;
    }

    int lastSound = 0;
    long remainder = target - m_Api.getTime();
    while (remainder > 0)
    {
      DommeSays(remainder + " seconds remaining.", 0, false, false);

      int soundByte = m_Api.getRandom(10);
      while (soundByte == lastSound)
      {
         soundByte = m_Api.getRandom(10);
      }
      lastSound = soundByte;

      String soundUrl = String.format("OralTrainer/encourage%02d.mp3", soundByte);
      m_Api.playSound(soundUrl);

      remainder = target - m_Api.getTime(); 
    }

    DommeSays("Up!", 0, false, false);
    m_Api.playSound("OralTrainer/up.mp3");

    def options = ["A- Down the whole time",
       "B- One or Two attempts",
       "C- Three or Four attempts",
       "D- Five attempts",
       "F- Total Failure"];

    int selection = m_Api.getSelectedValue("Now give yourself a grade.", options);

    selection -= BuyAGrade(selection);

    if (selection < 1)
    {
        int delta = m_EnduranceRating;
        DommeSays("I'm so proud of you. Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;

        m_EnduranceRating *= 1.20;

        if (selection < 0)
        {
          DommeSays("And since you're so eager to please, I'll make your next orgasm a few points closer.", 5, false, true);
          m_NextOrgasm -= 0.5 * delta;
        }
    }
    else if (selection== 1)
    {
        int delta = m_EnduranceRating / 2.0;
        DommeSays("Not bad. Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;

        m_EnduranceRating *= 1.10;
    }
    else if (selection == 2)
    {
        int delta = m_EnduranceRating / 4.0;
        DommeSays("Barely adequate. Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;
    }
    else if (selection == 3)
    {
        DommeSays("Your behavior is unacceptable! No points.", 5, false, true);

        m_EnduranceRating *= 0.90;
    }
    else
    {
        int delta = m_EnduranceRating;
        DommeSays("You did so badly that I'm actually going to punish you. Lets add " + delta + " points to your next orgasm.", 5, false, true);
        m_NextOrgasm += delta;

        m_EnduranceRating *= 0.80;
    }
  }

  private static void SpeedSession()
  {
    DommeSays("Lets work on your speed.", 3, false, true);

    if (m_SpeedRating < 1)
    {
      SpeedTest();
    }

    int sessionTime = 1;
    int endTime = m_Api.getTime() + sessionTime * 60;
    DommeSays("Up and down on my command...", 3, false, true);

    DommeSays("Open Wide!", 0, false, false);
    m_Api.playSound("OralTrainer/open.mp3");

    Wait(2, false);

    int target = Math.ceil(sessionTime * m_SpeedRating * 0.95 * m_Warmup);
    if (m_Warmup < 0.99)
    {
      m_Warmup += 0.25;
    }

    int rateMs = 60000 * sessionTime / (double)(target * 2.0); // 60000 ms per minute, divide by two since there are two beats
    int remainder = endTime - m_Api.getTime();
    while (remainder > 0)
    {
        DommeSays("Down!\n\n" + remainder + " seconds remaining.", 0, true, false);
	m_Api.playBackgroundSound("OralTrainer/down.mp3");
        ShortWait(rateMs);

        DommeSays("Up!\n\n" + remainder + " seconds remaining.", 0, false, false);
        m_Api.playBackgroundSound("OralTrainer/up.mp3");
        ShortWait(rateMs);

        remainder = endTime - m_Api.getTime();
    }

    def options = ["A- " + target + " or more.",
       "B- " + (int)Math.ceil(0.9 * target) + " or more.",
       "C- " + (int)Math.ceil(0.8 * target) + " or more.",
       "D- " + (int)Math.ceil(0.7 * target) + " or more.",
       "F- Anything less"];

    DommeSays("", 0, false, false);
    int selection = m_Api.getSelectedValue("Now how many FULL thrusts did you do.", options);

    selection -= BuyAGrade(selection);

    if (selection < 1)
    {
        int delta = m_SpeedRating;
        DommeSays("I'm so proud of you. Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;

        m_SpeedRating *= 1.5;

        if (selection < 0)
        {
          DommeSays("And since you're so eager to please, I'll make your next orgasm a few points closer.", 5, false, true);
          m_NextOrgasm -= 0.5 * delta;
        }
    }
    else if (selection == 1)
    {
        int delta = 0.5 * m_SpeedRating;
        DommeSays("Not bad. Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;

        m_SpeedRating *= 1.25;
    }
    else if (selection == 2)
    {
        int delta = 0.25 * m_SpeedRating;
        DommeSays("Barely adequate. Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;
    }
    else if (selection == 3)
    {
        DommeSays("Your behavior is unacceptable! No points.", 5, false, true);

        m_SpeedRating *= 0.90;
    }
    else
    {
        int delta = m_SpeedRating;
        DommeSays("You did so badly that I'm actually going to punish you. Lets add " + delta + " points to your next orgasm.", 5, false, true);
        m_NextOrgasm += delta;

        m_SpeedRating *= 0.80;
    }
  }

  private static void LickSession()
  {
    DommeSays("Maybe an easy one this time. Let's have you lick up and down the shaft.", 5, false, true);
    DommeSays("One means put you tongue at the tip, ten means at the base.\n\n I'm sure you can fill in the rest.", 5, false, true);
    DommeSays("Your tongue never leaves the cock.", 3, false, true);

    if (m_LickRating < 1)
    {
        m_LickRating = 30;
    }

    int sessionTime = 1;
    int endTime = m_Api.getTime() + sessionTime * 60;
    int target = Math.ceil(sessionTime * m_LickRating);
    int rateMs = 60000 * sessionTime / (double)target;
    int remainder = endTime - m_Api.getTime();
    int lastPosition = 1;
    int currentPosition = 1; 
    while (remainder > 0)
    {
        lastPosition = currentPosition;
        DommeSays("Position " + currentPosition + "!\n\n" + remainder + " seconds remaining.", 0, false, false);
        String soundUrl = String.format("OralTrainer/pos%02d.mp3", currentPosition);
	m_Api.playBackgroundSound(soundUrl);
        ShortWait(rateMs);

        while(currentPosition == lastPosition)
        {
            currentPosition = m_Api.getRandom(10) + 1;
        }

        remainder = endTime - m_Api.getTime();
    }

    def options = ["A- " + (int)Math.ceil(0.1 * target) + " or less.",
       "B- " + (int)Math.ceil(0.2 * target) + " or less.",
       "C- " + (int)Math.ceil(0.3 * target) + " or less.",
       "D- " + (int)Math.ceil(0.4 * target) + " or less.",
       "F- Too many times"];

    DommeSays("", 0, false, false);
    int selection = m_Api.getSelectedValue("How many times did you mess up (leave the dildo or go to the wrong position).", options);

    selection -= BuyAGrade(selection);

    if (selection < 1)
    {
        int delta = m_LickRating * 0.5;
        DommeSays("Good " + GetPronoun() + ". Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;

        m_LickRating *= 1.5;

        if (selection < 0)
        {
          DommeSays("And since you're so eager to please, I'll make your next orgasm a few points closer.", 5, false, true);
          m_NextOrgasm -= 0.5 * delta;
        }
    }
    else if (selection == 1)
    {
        int delta = 0.25 * m_LickRating;
        DommeSays("Not bad. Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;

        m_LickRating *= 1.25;
    }
    else if (selection == 2)
    {
        int delta = 0.1 * m_LickRating;
        DommeSays("Barely adequate. Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;
    }
    else if (selection == 3)
    {
        DommeSays("Your behavior is unacceptable! No points.", 5, false, true);

        m_LickRating *= 0.9;
    }
    else
    {
        int delta = m_LickRating;
        DommeSays("You did so badly that I'm actually going to punish you. Lets add " + delta + " points to your next orgasm.", 5, false, true);
        m_NextOrgasm += delta;

        m_LickRating *= 0.8;
    }

    if (m_LickRating > 120)
    {
      m_LickRating = 120;
    }
  }

  private static void SkillSession()
  {
    DommeSays("Maybe a tricky one this time. Let's have you take the dildo in your throat just to the specified position.", 5, false, true);
    DommeSays("One means kiss the tip, ten means means your lips press against the base.\n\n I'm sure you can fill in the rest.", 5, false, true);
    DommeSays("Your lips always circle the shaft.", 3, false, true);

    if (m_SpeedRating < 1)
    {
      SpeedTest();
    }

    int sessionTime = 1;
    int endTime = m_Api.getTime() + sessionTime * 60;
    int target = Math.ceil(sessionTime * m_SpeedRating) / 2.0 * m_Warmup;
    if (m_Warmup < 0.99)
    {
      m_Warmup += 0.25;
    }

    // Going too slow is hard too
    target = Math.max(10, target);

    int rateMs = 60000 * sessionTime / (double)target;
    int remainder = endTime - m_Api.getTime();
    int lastPosition = 10;
    int currentPosition = 10; 
    while (remainder > 0)
    {
        lastPosition = currentPosition;
        DommeSays("Position " + currentPosition + "!\n\n" + remainder + " seconds remaining.", 0, false, false);
        String soundUrl = String.format("OralTrainer/pos%02d.mp3", currentPosition);
	m_Api.playBackgroundSound(soundUrl);
        ShortWait(rateMs);

        while(currentPosition == lastPosition)
        {
            currentPosition = m_Api.getRandom(10) + 1;
        }

        remainder = endTime - m_Api.getTime();
    }

    def options = ["A- " + (int)Math.ceil(0.1 * target) + " or less.",
       "B- " + (int)Math.ceil(0.2 * target) + " or less.",
       "C- " + (int)Math.ceil(0.3 * target) + " or less.",
       "D- " + (int)Math.ceil(0.4 * target) + " or less.",
       "F- Too many times"];

    DommeSays("", 0, false, false);
    int selection = m_Api.getSelectedValue("How many times did you mess up (leave the dildo or go to the wrong position).", options);

    selection -= BuyAGrade(selection);

    if (selection < 1)
    {
        int delta = m_SpeedRating;
        DommeSays("Good " + GetPronoun() + ". Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;

        m_SpeedRating *= 1.25;

        if (selection < 0)
        {
          DommeSays("And since you're so eager to please, I'll make your next orgasm a few points closer.", 5, false, true);
          m_NextOrgasm -= 0.5 * delta;
        }
    }
    else if (selection == 1)
    {
        int delta = 0.5 * m_SpeedRating;
        DommeSays("Not bad. Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;

        m_SpeedRating *= 1.10;
    }
    else if (selection == 2)
    {
        int delta = 0.25 * m_SpeedRating;
        DommeSays("Barely adequate. Lets add " + delta + " points to your score.", 5, false, true);
        m_CurrentScore += delta;
    }
    else if (selection == 3)
    {
        DommeSays("Your behavior is unacceptable! No points.", 5, false, true);

        m_SpeedRating *= 0.9;
    }
    else
    {
        int delta = m_SpeedRating;
        DommeSays("You did so badly that I'm actually going to punish you. Lets add " + delta + " points to your next orgasm.", 5, false, true);
        m_NextOrgasm += delta;

        m_SpeedRating *= 0.8;
    }
  }

  private static int BuyAGrade(int original)
  {
    String prompt = "Pathetic. I'm going to offer you this once chance to redeem yourself.";
    switch(original)
    {
    case 0: // A
      prompt = "You know... good " + GetPronoun() + "s always try for extra credit...";
      break;
    case 1: // B
      prompt = "I would so like to give you a better grade. Maybe we can work something out?";
      break;
    case 2: // C
      prompt = "A poor performance. But perhaps there's a way to bump up that grade?";
      break;
    case 3: // D
    case 4: // F
    default:
      break;
    }

    def options = [
      "okay",
      "nope"
    ];

    int selection = m_Api.getSelectedValue(prompt, options);
    if (selection == 0)
    {
      if (m_IsMale)
      {
        return BuyAGradeMale();
      }
      else
      {
        return BuyAGradeFemale();
      }
    }
    else if (original > 2)
    {
      int delta = m_NextOrgasm * (0.05 * original);
      DommeSays("Extra punishment for refusing my generosity! " + delta + " more points before your next orgasm.", 6, false, true);
      m_NextOrgasm += delta;
    }

    return 0;
  }

  private static int BuyAGradeMale()
  {
    def options = [
      "I changed my mind.", //0
      "I'll put a clothespin on each nipple.", // 1
      "I'll put a clothespin on my cock.", // 2
      "I'll slap my balls 10 times.", // 3
      "I'll slap my balls 50 times.", //4
      "I'll spank myself 10 times on each butt cheek.", // 5
      "I'll spank myself 25 times on each butt cheek.", // 6
      "I'll write cock whore across my chest in sharpie.", // 7
      "I'll write cock whore across my face in sharpie.", // 8
      "I'll put a rubber band around my cock.", // 9
      "I'll fuck the dildo anally.", // 10
      "I'll wear a butt plug for the rest of the session.", // 11
    ];

    int selection = m_Api.getSelectedValue("What would you like to offer?", options);
    
    switch (selection)
    {
    case 0:
      DommeSays("Minus one letter grade for wasting my time.", 3, false, true);
      return -1;
    case 1:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Naturally, they stay there for the rest of the session.", ["Done"]);
      return 1;
    case 2:
      DommeSays("Naturally, they stay there for the rest of the session... even if you earn an orgasm.", 3, false, true);

      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("I hope your cock is completely covered when you finally earn permission to orgasm!", ["Done"]);
      return 1;
    case 3:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Do it HARD. This is a punishment!", ["Done"]);
      return 1;
    case 4:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Do it HARD. Lets add a new meaning to the term 'blue balls'!", ["Done"]);
      return 2;
    case 5:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Do it HARD. This is a punishment!", ["Done"]);
      return 1;
    case 6:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Do it HARD. I want to see hand-prints!", ["Done"]);
      return 2;
    case 7:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Consider yourself lucky its not a tattoo!", ["Done"]);
      return 1;
    case 8:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Cock across the forehead. Whore across the face- your lips make the 'O'.", ["Done"]);
      return 1;
    case 9:
      DommeSays("Make it tight! I want it purple and numb at the end of the session.", 5, false, true);

      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("I almost hope you earn an orgasm when your cock is all numb!", ["Done"]);
      return 1;
    case 10:
      DommeSays("Lube and straddle the dildo....", 0, false, false);
      Wait(15, true);

      DommeSays("Now fuck yourself. No Cumming!", 0, false, false);
      m_Api.playBackgroundSound("OralTrainer/wait.mp3");
      Wait(60, true);

      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("I hope you're clean, because its going right back in your mouth.", ["Done"]);
      return 2;
    case 11:
      DommeSays("Lube the butt plug and lie on your back with the plug against your hole....", 0, false, false);
      Wait(15, true);

      DommeSays("Now work it all the way in and out of your hole. Savage that ass for me!", 0, false, false);
      m_Api.playBackgroundSound("OralTrainer/wait.mp3");
      Wait(60, true);

      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Now shove it in deep and leave it!", ["Done"]);
      return 2;
    default:
      DommeSays("What are you talking about? Minus one grade.", 3, false, true);
      return -1;
    }
  }

  private static int BuyAGradeFemale()
  {
    def options = [
      "I changed my mind.", // 0
      "I'll put a clothespin on each breast.", // 1
      "I'll put a clothespin on each pussy lip.", // 2
      "I'll slap my tits 10 times each.", // 3
      "I'll slap my tits 25 times each.", // 4
      "I'll spank myself 10 times on each butt cheek.", // 5
      "I'll spank myself 25 times on each butt cheek.", // 6
      "I'll write cock whore across my chest in sharpie.", // 7
      "I'll write cock whore across my face in sharpie.", // 8
      "I'll put a rubber band around each breast.", // 9
      "I'll fuck the dildo anally.", // 10
      "I'll wear a butt plug for the rest of the session.", // 11
    ];

    int selection = m_Api.getSelectedValue("What would you like to offer?", options);

switch (selection)
    {
    case 0:
      DommeSays("Minus one letter grade for wasting my time.", 3, false, true);
      return -1;
    case 1:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Naturally, they stay there for the rest of the session.", ["Done"]);
      return 1;
    case 2:
      DommeSays("Naturally, they stay there for the rest of the session... even if you earn an orgasm.", 5, false, true);

      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("I almost hope you earn an orgasm and have to torture yourself to cum!", ["Done"]);
      return 1;
    case 3:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Do it HARD. This is a punishment!", 3, false);
      return 1;
    case 4:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Do it HARD. I want them to ache when the bra goes back on!", ["Done"]);
      return 2;
    case 5:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Do it HARD. This is a punishment!", ["Done"]);
      return 1;
    case 6:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Do it HARD. I want to see hand-prints!", ["Done"]);
      return 2;
    case 7:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Consider yourself lucky its not a tattoo!", ["Done"]);
      return 1;
    case 8:
      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Cock across the forehead. Whore across the face- your lips make the 'O'.", ["Done"]);
      return 1;
    case 9:
      DommeSays("Make it tight! I want them purple and numb at the end of the session.", 5, false, true);

      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("At least tit slaps wont hurt as much when they finally go numb!", ["Done"]);
      return 1;
    case 10:
      DommeSays("Lube and straddle the dildo....", 0, false, false);
      Wait(15, true);

      DommeSays("Now fuck yourself. No Cumming!", 0, false, false);
      m_Api.playBackgroundSound("OralTrainer/wait.mp3");
      Wait(60, true);

      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("I hope you're clean, because its going right back in your mouth.", ["Done"]);

      DommeSays("Do be careful if you put this in your pussy later.", 5, false, true);
      return 2;
    case 11:
      DommeSays("Lube the butt plug and lie on your back with the plug against your hole....", 0, false, false);
      Wait(15, true);

      DommeSays("Now work it all the way in and out of your hole. Savage that ass for me!", 0, false, false);
      m_Api.playBackgroundSound("OralTrainer/wait.mp3");
      Wait(60, true);

      DommeSays("", 0, false, false);
      m_Api.getSelectedValue("Now shove it in deep and leave it!", ["Done"]);
      return 2;
    default:
      DommeSays("What are you talking about? Minus one grade.", 3, false, true);
      return -1;
    }
  }

  public static boolean DebugMode()
  {
    DommeSays("Cheater, cheater, filthy cock eater. I can't stop you, but I don't have to be nice about it.", 5, false, true);

    def options = ["Continue Normally", // 0
       "Reset User Data", // 1
       "Set Score", // 2
       "Set Next Orgasm", // 3
       "Set Last Session", // 4
       "Set Endurance Rating", // 5
       "Set Speed Rating", // 6
       "Set Gender", // 7
       "Set Name", // 8
       "Endurance Test", // 9
       "Speed Test", // 10
       "Intro Sequence", // 11
       "Return Sequence", // 12
       "Endurance Session", // 13
       "Speed Session", // 14
       "Lick Session", // 15
       "Skill Session", // 16
       "Session Loop", // 17
       "Orgasm Sequence", // 18
       "Buy a Grade", // 19
       "Exit Script", // 20
       
    ];

    while (true)
    {
      int selection = m_Api.getSelectedValue("What now?", options);
      switch (selection)
      {
      case 0:
         DommeSays("Normally... right.", 3, false, true);
         return true;
      case 1:
         DommeSays("Forgetting everything about you...", 3, false, true);
         ClearUserData();
         break;
      case 2:
         m_CurrentScore = m_Api.getInteger("What is your score?", m_CurrentScore);
         break;
      case 3:
         m_NextOrgasm = m_Api.getInteger("How many points for your next orgasm?", m_NextOrgasm);
         break;
      case 4:
         m_LastSession = Long.parseLong(m_Api.getString("When was your last session?", m_LastSession));
         break;
      case 5:
         m_EnduranceRating = m_Api.getInteger("How long can you stay down at a time (seconds)?", m_EnduranceRating);
         break;
      case 6:
         m_SpeedRating = m_Api.getInteger("How many thrusts can you take per minute?", m_SpeedRating);
         break;
      case 7:
         m_IsMale = m_Api.getSelectedValue("What are you? I can't tell...?", ["male", "female"]) == 0;
         break;
      case 8:
         m_Name = m_Api.getString("So we're roleplaying now? What shall I call you then?", m_Name);
         break;
      case 9:
         DommeSays("Lets figure out your real endurance rating...", 3, false, true);
         EnduranceTest();
         break;
      case 10:
         DommeSays("Lets figure out your real speed rating...", 3, false, true);
         SpeedTest();
         break;
      case 11:
         DommeSays("Lets test out the newbie intro...", 3, false, true);
         IntroSequence();
         break;
      case 12:
         DommeSays("Lets test out the returning player intro...", 3, false, true);
         ReturnSequence();
         break;
      case 13:
         EnduranceSession();
         break;
      case 14:
         SpeedSession();
         break;
      case 15:
         LickSession();
         break;
      case 16:
         SkillSession();
         break;
      case 17:
         DommeSays("Lets try the main game loop...", 3, false, true);
         GameSequence();
         break;
      case 18:
         DommeSays("Now THAT is just cheating...", 3, false, true);
         OrgasmSequence();
         break;
      case 19:
         C19 : {
             DommeSays("Practice makes perfect!", 3, false, true);
             def grades = [
               "A",
               "B",
               "C",
               "D",
               "F"
             ];

             DommeSays("", 0, false, false);
             int grade = m_Api.getSelectedValue("What is the starting grade?", grades);
             int newGrade = grade - BuyAGrade(grade);
             String newGradeStr = newGrade < 0 ? "A+" : (newGrade > 4 ? "F-" : grades[newGrade]);
             DommeSays(String.format("You went from grade %s to grade %s.", grades[grade], newGradeStr), 5, false, true);
         }
         break;
      case 20:
         DommeSays("You are the weakest link... goodbye.", 3, false, true);
         return false;
      default:
         break;
      }
    }

    return false;
  }

  private static String GetPronoun()
  {
    if (m_IsMale)
    {
      return "boy";
    }
    else
    {
      return "girl";
    }
  }

  private static boolean m_IsMale = false;

  private static String m_Name = "";

  private static def m_Api = null;

  private static long m_LastSession = 0;

  private static int m_NextOrgasm = 0;

  private static int m_CurrentScore = 0;

  private static int m_EnduranceRating = 0;

  private static int m_SpeedRating = 0;

  private static int m_LickRating = 0;

  private static int m_LastPicture = 0;

  private static boolean m_IsDebug = false;

  private static double m_Warmup = 0.5;

  private static def m_WaitFxn = null;
}
